/*

This file is (C) Copyright 2002 Erik S. Johansen <einride@einride.org>, all 
rights reserved.

If you work for or represent an organisation, agency or corporation that is 
wholly or partially funded by a government -OR- you are in any way associated 
with the company Cyber Operations, the following licence applies.

You may not use this file or any derivatives of it for any purpose whatsoever,
including reading beyond this comment, redistribution or compiling, without 
explicit written permission from the copyright holder. 

If you do not work for or represent an organisation, agency or corporation 
that is wholly or partially funded by a government -AND- you are not in any 
way associated with the company Cyber Operations, the GNU General Public 
License, as described in the accompanying file GPL, applies.

*/

#include "opcode.h"

struct instruction instructions[] = {
  {"int3",                 "CC"},
  {"int",                  "CD?1"},
  {"test eax, eax",        "85C0"},
  {"mov eax, imm",         "B8?4"},
  {"mov ebx, imm",         "BB?4"},
  {"mov ecx, imm",         "B9?4"},
  {"mov edx, imm",         "BA?4"},
  {"mov eax, abs",         "B8?A"},
  {"mov ebx, abs",         "BB?A"},
  {"mov ecx, abs",         "B9?A"},
  {"mov edx, abs",         "BA?A"},
  {"jnz",                  "75?R"},
  {"push imm4",            "68?4"},
  {"push abs",             "68?A"},
  {"push eax",             "50"},
  {"pop ebx",              "5b"},
  {"nop",                  "90"},
  {"mov ecx, esp",         "89E1"},
  {"mov ebp, eax",         "89C5"},
  {"cmp eax, imm",         "3D?4"},
  {"jl",                   "7C?R"},
  
  {NULL,                   NULL},
};


#define OPCODE_LABEL_LENGTH 64

#define CK_DATA 1
#define CK_ABSREF 2
#define CK_RELREF 3


struct code_entry {
  /* This codes offset after last calculation */
  unsigned long offset;                 

  /* This codes label */
  char label[OPCODE_LABEL_LENGTH];      

  /* What is this code really CK_DATA || CK_ABSREF || CK_RELREF */
  int kind;                            

  /* Actual data if kind == CK_DATA */
  unsigned char * data;     

  /* Data size. 1 for CK_RELREF, 4 for CK_ABSREF, varying for CK_DATA */
  int datasize;
  
  /* Which label do this refer to? Valid if kind == CK_RELREF || kind == CK_ABSREF */
  char refto[OPCODE_LABEL_LENGTH];
  
  /* Calculated offset if kind == CK_ABSREF */
  unsigned long absref;

  /* Calculated offset if kind == CK_RELREF */
  char relref;
};

struct code {
  int count, size;
  struct code_entry * code;
};


/* create a new opcode collection for use with code_* functions */
pcode code_create() {
  pcode res;
  res = malloc(sizeof(struct code));
  memset(res, 0, sizeof(struct code));
  return res;
}

/* Cleanup an opcode collection */
void code_kill(pcode pc) {
  int index;
  if (!pc)
    return;
  for (index=0; index<pc->count; index++) {
    if (pc->code[index].data)
      free(pc->code[index].data);
  }
  if (pc->code)
    free(pc->code);
  memset(pc, 0, sizeof(struct code));
  free(pc);
}


/* Internally used to create a new code_entry in a pcode */
struct code_entry * code_addentry(pcode pc, char * label) {
  struct code_entry * res;
  if (!pc) {
    fprintf(stderr, "Fatal: code_addentry called with NULL pcode\n");
    exit(1);
  }
  pc->code = realloc(pc->code, (pc->count + 1) * sizeof (struct code_entry));
  res = &pc->code[pc->count];
  memset(res, 0, sizeof(struct code_entry));
  if (label) {
    strncpy(res->label, label, sizeof(res->label));
    res->label[sizeof(res->label)-1] = 0;
  }
  pc->count++;
  return res;
}

/* Internally used to append a CK_ABSREF entry */
void code_addabsref(pcode pc, char * label, char * reflabel) {
  struct code_entry * ce;
  if (!reflabel) {
    fprintf(stderr, "Fatal: Attempted to add an absolute reference but no label specified\n");
    exit(1);
  }
  ce = code_addentry(pc, label);
  ce->kind = CK_ABSREF;
  strncpy(ce->refto, reflabel, sizeof(ce->refto));
  ce->refto[sizeof(ce->refto)-1] = 0;
  ce->datasize=4;
}

/* Internally used to append an CK_RELREF entry */
void code_addrelref(pcode pc, char * label, char * reflabel) {
  struct code_entry * ce;
  if (!reflabel) {
    fprintf(stderr, "Fatal: Attempted to add a relative reference but no label specified\n");
    exit(1);
  }
  ce = code_addentry(pc, label);
  ce->kind = CK_RELREF;
  strncpy(ce->refto, reflabel, sizeof(ce->refto));
  ce->refto[sizeof(ce->refto)-1] = 0;
  ce->datasize=1;  
}

/* Internally used to find a labels corresponding code_entry */ 
struct code_entry * code_findlabel(pcode pc, char * label) {
  int i;
  if (!label || !label[0]) {
    fprintf(stderr, "Fatal: Attempted to scan pcode for a NULL label\n");
    exit(1);
  }
  if (!pc) {
    fprintf(stderr, "Fatal: Attempted to scan a NULL pcode for label '%s'\n", label);
    exit(1);
  }
  for (i=0;i<pc->count;i++)
    if (!strcmp(pc->code[i].label, label))
      return &pc->code[i];
  return NULL;
}

/* Append an instruction to an opcode collection */
void code_addins(pcode pc, char * label, char * insname, ...) {
  int insindex;
  va_list va;
  char * ins;
  char insout[128];
  int insoutp;

  /* Find the entry in the 'instructions' array */
  for (insindex=0; instructions[insindex].name; insindex++)
    if (!strcmp(insname, instructions[insindex].name))
      break;
  if (!instructions[insindex].name) {
    fprintf(stderr, "Fatal: Instruction '%s' was not found in instruction list\n", insname);
    exit(1);
  }

  /* Scan and decode the opcode spec */
  va_start(va, insname);
  ins = instructions[insindex].opcode;
  insoutp=0;
  memset(insout, 0, sizeof(insout));
  while (*ins) {
    if (*ins == '?') {
      ins++;
      if (*ins=='1') {
	* (unsigned char *) &insout[insoutp] = va_arg(va, unsigned char);
	insoutp++;
      } else if (*ins=='2') {
	* (unsigned short *) &insout[insoutp] = va_arg(va, unsigned short);
	insoutp += 2;
      } else if (*ins=='4') {
	* (unsigned int *) &insout[insoutp] = va_arg(va, unsigned int);
	insoutp += 4;
      } else if (*ins=='A') {
	if (insoutp) {
	  code_addraw(pc, label, &insout, insoutp);
	  label = NULL;
	  insoutp=0;
	  memset(&insout, 0, sizeof(insout));
	};
	code_addabsref(pc, label, va_arg(va, char *));
	label = NULL;
      } else if (*ins=='R') {
	if (insoutp) {
	  code_addraw(pc, label, &insout, insoutp);
	  label = NULL;
	  insoutp=0;
	  memset(&insout, 0, sizeof(insout));
	};
	code_addrelref(pc, label, va_arg(va, char *));
	label = NULL;
      } else {
	fprintf(stderr, "Fatal: Instruction '%s' has an invalid opcode definition\n", insname);
	exit(1);
      }
      ins++;
    } else {
      char tmp[3], *p;
      unsigned long u;

      strncpy(tmp, ins, 3);
      tmp[2] = 0;
      u = strtoul(tmp, &p, 16);
      if (*p != 0) {
	fprintf(stderr, "Fatal: Instruction '%s' has an invalid opcode definition\n", insname);
	exit(1);
      }
      insout[insoutp++] = (unsigned char) (u & 0xFF);
      ins += 2;
    }
  }
  va_end(va);
  if (insoutp)
    code_addraw(pc, label, &insout, insoutp);
}

/* Append data to an opcode collection */
void code_addraw(pcode pc, char * label, void * data, int datalen) {
  struct code_entry * ce;
  if (!data || !datalen) {
    fprintf(stderr, "Fatal: Attempted to add NULL raw data\n");
    exit(1);
  }
  ce = code_addentry(pc, label);
  ce->kind = CK_DATA;
  ce->data = malloc(datalen);
  memcpy(ce->data, data, datalen);
  ce->datasize = datalen;
}

/* Append an immediate value to an opcode collection */
void code_addbyte(pcode pc, char * label, BYTE v) {
  code_addraw(pc, label, &v, 1);
}

void code_addword(pcode pc, char * label, WORD v) {
  code_addraw(pc, label, &v, 2);
}

void code_adddword(pcode pc, char * label, DWORD v) {
  code_addraw(pc, label, &v, 4);
}

/* Calculate the buffer size for an opcode collection */
int code_size(pcode pc) {
  int size=0, i;
  if (!pc)
    return 0;
  for (i=0; i<pc->count;i++)
    size += pc->code[i].datasize;
  return size;
}

/* Write opcodes adjusted for specified base address to buffer */
void code_write(pcode pc, DWORD base, void * buf) {
  int curofs=base, i, bpos;
  char * buffer = buf;
  struct code_entry * ce;

  if (!pc)
    return;

  /* Pass 1 sets all offsets */
  for (i=0;i<pc->count;i++) {
    pc->code[i].offset = curofs;
    curofs += pc->code[i].datasize;
  }

  /* Pass 2 resolves all references */
  for (i=0; i<pc->count; i++) 
    if ((pc->code[i].kind == CK_ABSREF) || (pc->code[i].kind == CK_RELREF)) {
      ce = code_findlabel(pc, pc->code[i].refto);
      if (!ce) {
	fprintf(stderr, "Fatal: Unresolved reference to '%s' in pcode\n", pc->code[i].refto);
	exit(1);
      }
      if (pc->code[i].kind == CK_ABSREF) {
	pc->code[i].absref = ce->offset;
      } else {
	int rel = ce->offset - ((int) pc->code[i].offset + 1);
	if ((rel & 0xFF) != rel) {
	  fprintf(stderr, "Fatal: Relative reference to '%s' too high\n", ce->label);
	  exit(1);
	}
	pc->code[i].relref = (char) rel;
      }
    }
  
  /* Pass 3 writes the buffer */
  
  for (i=0, bpos=0; i<pc->count; i++) {
    if (pc->code[i].kind == CK_DATA) {
      memcpy(&buffer[bpos], pc->code[i].data, pc->code[i].datasize);
      bpos += pc->code[i].datasize;
    } else if (pc->code[i].kind == CK_ABSREF) {
      * (unsigned long *) &buffer[bpos] = pc->code[i].absref;
      bpos += 4;
    } else if (pc->code[i].kind == CK_RELREF) {
      buffer[bpos] = pc->code[i].relref;
      bpos += 1;
    } else {
      fprintf(stderr, "Fatal: Unknown code entry kind. Shouldn't happen\n");
      exit(1);
    }
  }
}


