/*

This file is (C) Copyright 2002 Erik S. Johansen <einride@einride.org>, all 
rights reserved.

If you work for or represent an organisation, agency or corporation that is 
wholly or partially funded by a government -OR- you are in any way associated 
with the company Cyber Operations, the following licence applies.

You may not use this file or any derivatives of it for any purpose whatsoever,
including reading beyond this comment, redistribution or compiling, without 
explicit written permission from the copyright holder. 

If you do not work for or represent an organisation, agency or corporation 
that is wholly or partially funded by a government -AND- you are not in any 
way associated with the company Cyber Operations, the GNU General Public 
License, as described in the accompanying file GPL, applies.

*/

#include "trace.h"

/* 
 * Internal func used to check whether a proc is in a state where
 * it can get ptrace requests. Will wait blocking if needed.
 */
int trace_canptrace(pprocess proc) {
  if (!proc) {
    errno = EFAULT;
    return 0;
  }
  if (!(proc->flags & TRACE_ATTACHED)) {
    errno = EBUSY;
    return 0;
  }
  if (!(proc->flags & TRACE_WAITED)) {
    kill(proc->pid, SIGCHLD);
    if (!trace_wait(proc, -1))
      return 1;
    else 
      return 0;
  }
  return 1;
}

pprocess trace_attach(pid_t pid) {
  pprocess proc;
  errno = 0;
  if (ptrace(PT_ATTACH, pid, 0, 0))
    return NULL;
  proc=malloc(sizeof(struct process));
  memset(proc, 0, sizeof(struct process));
  proc->pid = pid;
  proc->flags = TRACE_ATTACHED;
  return proc;
}

int trace_detach(pprocess proc) {
  if (!trace_canptrace(proc))
    return -1;
  if (!(proc->flags & TRACE_EXITED))
    if (ptrace(PT_DETACH, proc->pid, (caddr_t) 1, 0)) {
      if (errno==ESRCH)
	proc->flags |= TRACE_EXITED;
      return -1;
    }
  memset(proc, 0, sizeof(struct process));
  free(proc);
  return 0;
}

int trace_step(pprocess proc) {
  if (!trace_canptrace(proc))
    return -1;
  if (!(proc->flags & TRACE_EXITED)) {
    errno = 0;
    if (ptrace(PT_STEP, proc->pid, (caddr_t) 1, 0)) {
      if (errno==ESRCH)
	proc->flags |= TRACE_EXITED;
      return -1;
    }
    proc->flags = TRACE_ATTACHED;
  } else {
    errno = ESRCH;
    return -1;
  }
  return 0;
}

int trace_write(pprocess proc, unsigned long address, void * data, int datalen) {
  int prefix, raddr, rlen, saddr, i;
  unsigned char * buf;
  if (!trace_canptrace(proc))
    return -1;

  /* dword align start address */
  prefix = (address & 3);
  raddr = address & ~3;
  rlen = (datalen + prefix + 3) & ~3;
  /*  suffix = (rlen - prefix - datalen); */
  saddr = (raddr + prefix + datalen);

  buf = malloc(rlen);
  if (prefix) {
    if (trace_read(proc, raddr, buf, 4)) {
      free(buf);
      return -1;
    }
  }
  if (rlen - prefix - datalen) {
    if (trace_read(proc, saddr, &buf[rlen-4], 4)) {
      free(buf);
      return -1;
    }
  }
  memcpy(&buf[prefix], data, datalen);
  for (i=0;i<rlen;i+=4) {
    if (ptrace(PT_WRITEDATA, proc->pid, address+i, * (unsigned int *) &buf[i])) {
      if (errno==ESRCH)
	proc->flags |= TRACE_EXITED;
      free(buf);
      return -1;
    }
  }
  free(buf);
  return 0;
}

/* 
 * Read from a process's address space
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */
int trace_read(pprocess proc, unsigned long address, void * data, int datalen) {
  int prefix, raddr, rlen, saddr, res, i;
  unsigned char * buf;
  if (!trace_canptrace(proc))
    return -1;

  /* dword align start address */
  prefix = (address & 3);
  raddr = address & ~3;
  rlen = (datalen + prefix + 3) & ~3;
  /*  suffix = (rlen - prefix - datalen); */
  saddr = (raddr + prefix + datalen);

  buf = malloc(rlen);
  for (i=0;i<rlen;i+=4) {
    errno = 0;
    if (((res=ptrace(PT_READDATA, proc->pid, raddr + i, 0))==-1) && (errno)) {
      if (errno==ESRCH)
	proc->flags |= TRACE_EXITED;
      free(buf);
      return -1;
    }
    * (int *) &buf[i] = res;
  }
  memcpy(data, &buf[prefix], datalen);
  free(buf);
  return 0;
}

/* Set a process's registers
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */
int trace_setregisters(pprocess proc, struct registers * regs) {
  errno = 0;
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (EAX*4), regs->eax)) {
    if (errno==ESRCH)
      proc->flags |= TRACE_EXITED;
    return -1;
  }
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (EBX*4), regs->ebx))
    return -1;
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (ECX*4), regs->ecx))
    return -1;
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (EDX*4), regs->edx))
    return -1;
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (ESI*4), regs->esi))
    return -1;
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (EDI*4), regs->edi))
    return -1;
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (EBP*4), regs->ebp))
    return -1;
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (UESP*4), regs->esp))
    return -1;
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (EIP*4), regs->eip))
    return -1;
  if (ptrace(PT_WRITEREG, proc->pid, (caddr_t) (EFL*4), regs->eflags))
    return -1;
  return 0;
}

/* Get a process's registers
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */
int trace_getregisters(pprocess proc, struct registers * regs) {
  errno = 0;
  regs->oeax = ptrace(PT_READREG, proc->pid, (caddr_t) (ORIG_EAX*4), 0);
  if (errno) {
    if (errno==ESRCH)
      proc->flags |= TRACE_EXITED;
    return errno;
  }
  regs->eax = ptrace(PT_READREG, proc->pid, (caddr_t) (EAX*4), 0);
  if (errno)
    return errno;
  regs->ebx = ptrace(PT_READREG, proc->pid, (caddr_t) (EBX*4), 0);
  if (errno)
    return errno;
  regs->ecx = ptrace(PT_READREG, proc->pid, (caddr_t) (ECX*4), 0);
  if (errno)
    return errno;
  regs->edx = ptrace(PT_READREG, proc->pid, (caddr_t) (EDX*4), 0);
  if (errno)
    return errno;
  regs->esi = ptrace(PT_READREG, proc->pid, (caddr_t) (ESI*4), 0);
  if (errno)
    return errno;
  regs->edi = ptrace(PT_READREG, proc->pid, (caddr_t) (EDI*4), 0);
  if (errno)
    return errno;
  regs->ebp = ptrace(PT_READREG, proc->pid, (caddr_t) (EBP*4), 0);
  if (errno)
    return errno;
  regs->esp = ptrace(PT_READREG, proc->pid, (caddr_t) (UESP*4), 0);
  if (errno)
    return errno;
  regs->eip = ptrace(PT_READREG, proc->pid, (caddr_t) (EIP*4), 0);
  if (errno)
    return errno;
  regs->eflags = ptrace(PT_READREG, proc->pid, (caddr_t) (EFL*4), 0);
  if (errno)
    return errno;
  return 0;
}

/* Let a process run until it gets a signal
 * if block is 1, the function will block waiting for the process
 * to receive a signal. If not, the process must be polled until
 * trace_wait returns non 0 before other trace_* functions
 * can be used.
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */ 
int trace_resume(pprocess proc, int block) {
  int sig;
  if (!trace_canptrace(proc))
    return -1;
  sig=trace_readsignal(proc);
  if ((sig==SIGSTOP) || (sig==SIGTRAP) || (sig==0))
    sig = SIGCHLD;
  if (ptrace(PT_CONT, proc->pid, 1, sig)) {
    if (errno==ESRCH)
      proc->flags |= TRACE_EXITED;
    return -1;
  }
  proc->flags &= ~(TRACE_WAITED);
  if (block)
    return trace_wait(proc, -1);
  else
    return 0;
}

int trace_resume_signal(pprocess proc, int block, int signal) {
  if (!trace_canptrace(proc))
    return -1;
  if (ptrace(PT_CONT, proc->pid, 1, signal)) {
    if (errno==ESRCH)
      proc->flags |= TRACE_EXITED;
    return -1;
  }
  proc->flags &= ~(TRACE_WAITED);
  if (block)
    return trace_wait(proc, -1);
  else
    return 0;
}

#ifdef __linux__
/* Let a process run until it gets a syscall or signal
 * if block is 1, the function will block waiting for the process
 * to receive a signal or syscall. If not, the process must be polled
 * until trace_wait returns non 0 before other trace_* functions
 * can be used.
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */ 
int trace_runtosyscall(pprocess proc, int block) {
  return -1;
}
#endif


void trace_sigalarm(int x) {
}

void trace_sigintr(int x) {
}

int trace_wait(pprocess proc, int waittime) {
  struct sigaction sa, old_sigintr, old_sigalarm;
  int res, err;
  if (!proc) {
    errno = EFAULT;
    return -1;
  }
  if (!proc->pid) {
    errno = EINVAL;
    return -1;
  }
  if (!(proc->flags & TRACE_ATTACHED)) {
    errno = EINVAL;
    return -1;
  }
  if (proc->flags & TRACE_WAITED)
    return 0;

  memset(&sa, 0, sizeof(sa));
  if (waittime>0) {
    sa.sa_handler = trace_sigintr;
    sigaction(SIGINT, &sa, &old_sigintr);
    sa.sa_handler = trace_sigalarm;
    sigaction(SIGALRM, &sa, &old_sigalarm);
    alarm(waittime);
    errno = 0;
    res = waitpid(proc->pid, &proc->stat, WUNTRACED);
    err = errno;
    alarm(0);
    sigaction(SIGALRM, &old_sigalarm, 0);
    sigaction(SIGINT, &old_sigintr, 0);
  } else if (waittime == 0) {
    errno=0;
    res=waitpid(proc->pid, &proc->stat, WNOHANG | WUNTRACED);
    err=errno;
  } else {
    sa.sa_handler = trace_sigintr;
    sigaction(SIGINT, &sa, &old_sigintr);
    errno=0;
    res=waitpid(proc->pid, &proc->stat, WUNTRACED);
    err=errno;
    sigaction(SIGINT, &old_sigintr, 0);
  }
  if (res < 0) {
    if (errno == ESRCH) {
      proc->flags = TRACE_EXITED | TRACE_ATTACHED;
      return 0;
    }
    return -1;
  } else if (!res) {
    if (errno==ESRCH)
      proc->flags |= TRACE_EXITED;
    return -1;
  } else {
    proc->flags |= TRACE_WAITED;
    if (WIFEXITED(proc->stat))
      proc->flags |= TRACE_EXITED;
    //    if (WSTOPSIG(proc->stat)!=SIGCHLD)
    //  printf("stopped: %s\n", strsignal(WSTOPSIG(proc->stat)));
    return 0;
  }
}

int trace_readsignal(pprocess proc) {
  if (!proc) {
    errno = EFAULT;
    return -1;
  }
  if (!proc->pid) {
    errno = EINVAL;
    return -1;
  }
  if (!(proc->flags & TRACE_ATTACHED)) {
    errno = EINVAL;
    return -1;
  }
  if (!(proc->flags & TRACE_WAITED))
    return 0;
  if (!WIFSTOPPED(proc->stat))
    return 0; 
  return WSTOPSIG(proc->stat);
}









