/*

This file is (C) Copyright 2002 Erik S. Johansen <einride@einride.org>, all 
rights reserved.

If you work for or represent an organisation, agency or corporation that is 
wholly or partially funded by a government -OR- you are in any way associated 
with the company Cyber Operations, the following licence applies.

You may not use this file or any derivatives of it for any purpose whatsoever,
including reading beyond this comment, redistribution or compiling, without 
explicit written permission from the copyright holder. 

If you do not work for or represent an organisation, agency or corporation 
that is wholly or partially funded by a government -AND- you are not in any 
way associated with the company Cyber Operations, the GNU General Public 
License, as described in the accompanying file GPL, applies.

*/

#ifndef _TRACE_H
#define _TRACE_H

/*
 * trace.h - Headerfile for system debugging wrapper functions
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ptrace.h>
#include <signal.h>

/* OS specific includes & defines */
#if defined(__linux__)

#include <asm/ptrace.h>

#define PT_ATTACH PTRACE_ATTACH
#define PT_READDATA PTRACE_PEEKDATA
#define PT_WRITEDATA PTRACE_POKEDATA
#define PT_STEP PTRACE_SINGLESTEP
#define PT_CONT PTRACE_CONT
#define PT_READREG PTRACE_PEEKUSER
#define PT_WRITEREG PTRACE_POKEUSER

#else
#include "_unknown_os_"
#endif



/* i386 registers */
struct registers {
  unsigned long eax, ebx, ecx, edx, esi, edi, ebp, esp, eip, eflags, oeax;
  unsigned short cs, ds, es, fs, gs, ss;
};


/* 
 * Internal process state struct.
 * This can be read, but must not be written
 * to directly.
 */
struct process {
  pid_t pid;
  struct registers cregs, nregs;
  unsigned long flags;
  int stat;
  void * data;
};
typedef struct process * pprocess;

/*
 * the process.flags bits
 */
#define TRACE_ATTACHED 0x00000001
#define TRACE_WAITED   0x00000002
#define TRACE_EXITED   0x00000004


/*
 * Macros to check state of a process. 
 */
#define TraceAttached(x) (x && x->pid && (x->flags & TRACE_ATTACHED))
#define TraceStopped(x) (TraceAttached(x) && (x->flags & TRACE_WAITED))
#define TraceExited(x) (x && x->pid && (x->flags & TRACE_EXITED))

/* Attach to the specified process, returning a pprocess which
 * is used for all subsequent trcae_* calls 
 * If an error occurs, NULL is returned and errno is set
*/
pprocess trace_attach(pid_t pid);

/* Detach from a process, letting it continue. The proc variable
 * will be invalid on return.
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */
int trace_detach(pprocess proc);


/* Single-step a process once
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */
int trace_step(pprocess proc);

/* Write to a process's address space
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */
int trace_write(pprocess proc, unsigned long address, void * data, int datalen);

/* Read from a process's address space
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */
int trace_read(pprocess proc, unsigned long address, void * data, int datalen);

/* Set a process's registers
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */
int trace_setregisters(pprocess proc, struct registers * regs);

/* Get a process's registers
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */
int trace_getregisters(pprocess proc, struct registers * regs);

/* Let a process run until it gets a signal
 * if block is 1, the function will block waiting for the process
 * to receive a signal. If not, the process must be polled until
 * trace_wait returns non 0 before other trace_* functions
 * can be used.
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */ 
int trace_resume(pprocess proc, int block);
int trace_resume_signal(pprocess proc, int block, int signal);

#ifdef __linux__
/* Let a process run until it gets a syscall or signal
 * if block is 1, the function will block waiting for the process
 * to receive a signal or syscall. If not, the process must be polled
 * until trace_wait returns non 0 before other trace_* functions
 * can be used.
 * Returns 0 if ok, errno is set and -1 returned if an error
 * occurs
 */ 
int trace_runtosyscall(pprocess proc, int block);
#endif


int trace_wait(pprocess proc, int waittime);

int trace_readsignal(pprocess proc);


#endif






